import pycurl
import os
import sys
import logging
import hashlib
import re

from io import BytesIO
from munch import Munch

from .utils import run, QUIET, MODEST, VERBOSE, download_file,\
        print_network_progress
from .errors import RpkgError

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

try:
    from urllib.parse import quote as url_quote
except ImportError:
    from urllib import quote as url_quote

log = logging.getLogger('__main__')

class LookasideCache(object):

    def __init__(self, pkg_path, download_url, upload_url,
                 client_cert=None, client_key=None,
                 ca_cert=None, verbosity=MODEST):
        self.pkg_path = pkg_path
        self.verbosity = verbosity
        self.download_url = download_url
        self.upload_url = upload_url
        self.client_cert = client_cert
        self.client_key = client_key
        self.ca_cert = ca_cert
        self.sources_filepath = os.path.join(self.pkg_path, 'sources')

    ######################## helpers ########################

    def get_download_url(self, repo_path, filename, hashtype, hash, **kwargs):
        subs = {
            'repo_path': repo_path.strip('/'),
            'filename': url_quote(filename),
            'hashtype': hashtype,
            'hash': hash,
        }
        subs.update(kwargs)
        try:
            result_url = self.download_url % subs
        except Exception as e:
            raise DownloadError(
                'Could not get download url from %s template. '
                'Original error: %s.' % (self.download_url, str(e)))

        return result_url

    def download(self, outfile, repo_path, filename, hashtype, hash, **kwargs):
        if os.path.exists(outfile):
            if self.file_is_valid(outfile, hashtype, hash):
                log.debug("File %s already downloaded.", outfile)
                return

        if not self.download_url:
            raise RpkgError('download_url has not been configured.')

        download_url = self.get_download_url(
            repo_path, filename, hashtype, hash, **kwargs)

        log.info("Downloading %s from %s at %s:",
                 filename, repo_path, urlparse(self.download_url).netloc)
        log.debug("Full url: %s", download_url)

        ca_cert = None
        if self.ca_cert:
            if os.path.exists(self.ca_cert):
                ca_cert = self.ca_cert
            else:
                log.warning("Missing CA certificate: %s", self.ca_cert)

        status = download_file(download_url, outfile, ca_cert=ca_cert)

        if status != 200:
            os.remove(outfile)
            log.info('Removed downloaded invalid file %s', outfile)
            raise RpkgError('Server returned status code %d.' % status)

        if not self.file_is_valid(outfile, hashtype, hash):
            raise RpkgError('%s failed checksum' % outfile)

    def hashsum(self, filepath, hashtype):
        h = hashlib.new(hashtype)
        outfile_h = open(filepath, 'rb')
        data = outfile_h.read(8192)

        while data:
            h.update(data)
            data = outfile_h.read(8192)

        return h.hexdigest()

    def file_is_valid(self, filepath, hashtype, hash):
        return self.hashsum(filepath, hashtype) == hash

    def read_source_entries(self):
        if not os.path.isfile(self.sources_filepath):
            return []

        with open(self.sources_filepath, 'r') as sources_file:
            source_lines = sources_file.readlines()

        source_entries = []
        for source_line in source_lines:
            m = re.match(r'(?P<hashtype>\w+)\s*\((?P<filename>.+)\)\s*=\s*(?P<hash>\w+)\s*',
                         source_line)

            if not m:
                raise RpkgError('Invalid entry in sources file: {0}'
                                .format(source_line.strip()))

            source_entries.append(Munch(
                filename=m.group('filename'),
                hashtype=m.group('hashtype').lower(),
                hash=m.group('hash')
            ))

        return source_entries

    def write_source_entries(self, source_entries, append=True):
        if append:
            current_source_entries = self.read_source_entries()
            sources_file = open(self.sources_filepath, 'a')
        else:
            current_source_entries = []
            sources_file = open(self.sources_filepath, 'w')

        for entry in source_entries:
            entry_exists = False
            for current_entry in current_source_entries:
                if entry == current_entry:
                    entry_exists = True
                    break

            if entry_exists:
                continue

            entry_line_template = '{hashtype} ({filename}) = {hash}\n'
            sources_file.write(entry_line_template.format(
                hashtype=entry.hashtype.upper(),
                filename=entry.filename,
                hash=entry.hash
            ))

        sources_file.close()

    def make_source_entry(self, filepath):
        return Munch(filename=os.path.basename(filepath),
                     hashtype='sha512',
                     hash=self.hashsum(filepath, 'sha512'))

    def add_upload_certs(self, c):
        if self.client_key:
            if os.path.exists(self.client_key):
                c.setopt(pycurl.SSLKEY, self.client_key)
            else:
                log.warning("Missing ssl key: %s", self.client_key)

        if self.client_cert:
            if os.path.exists(self.client_cert):
                c.setopt(pycurl.SSLCERT, self.client_cert)
            else:
                log.warning("Missing client certificate: %s", self.client_cert)

        if self.ca_cert:
            if os.path.exists(self.ca_cert):
                c.setopt(pycurl.CAINFO, self.ca_cert)
            else:
                log.warning("Missing CA certificate: %s", self.ca_cert)

    def source_uploaded(self, repo_path, filename, hashtype, hash):
        buf = BytesIO()

        c = pycurl.Curl()
        c.setopt(pycurl.URL, str(self.upload_url))
        c.setopt(pycurl.HTTPAUTH, pycurl.HTTPAUTH_GSSNEGOTIATE)
        c.setopt(pycurl.USERPWD, ':')
        c.setopt(pycurl.WRITEDATA, buf)
        c.setopt(pycurl.FOLLOWLOCATION, True)
        c.setopt(pycurl.MAXREDIRS, 1)
        c.setopt(pycurl.HTTPPOST, [
            ('filename', filename),
            ('name', repo_path),
            (hashtype+'sum', hash),
        ])
        self.add_upload_certs(c)

        try:
            c.perform()
            status = c.getinfo(pycurl.RESPONSE_CODE)
            output = buf.getvalue().decode('utf-8').strip()
        except Exception as e:
            raise RpkgError(e)
        finally:
            c.close()

        if status != 200:
            log.error(output)
            raise RpkgError('Server returned status code %d.' % status)
        else:
            log.debug(output)

        if output == 'Available':
            return True

        return False

    def upload_source(self, infile, repo_path, filename, hashtype, hash):
        if not self.upload_url:
            raise RpkgError('upload_url has not been configured.')

        log.info('Uploading %s to %s at %s:',
                 filename, repo_path, urlparse(self.upload_url).netloc)

        if self.source_uploaded(repo_path, filename, hashtype, hash):
            log.info('File %s already uploaded.', filename)
            return

        buf = BytesIO()

        c = pycurl.Curl()
        c.setopt(pycurl.URL, str(self.upload_url))
        c.setopt(pycurl.HTTPAUTH, pycurl.HTTPAUTH_GSSNEGOTIATE)
        c.setopt(pycurl.USERPWD, ':')
        c.setopt(pycurl.NOPROGRESS, False)
        c.setopt(pycurl.PROGRESSFUNCTION, print_network_progress)
        c.setopt(pycurl.WRITEDATA, buf)
        c.setopt(pycurl.FOLLOWLOCATION, True)
        c.setopt(pycurl.MAXREDIRS, 1)
        c.setopt(pycurl.HTTPPOST, [
            ('file', (pycurl.FORM_FILE, infile)),
            ('mtime', str(os.path.getmtime(infile))),
            ('name', repo_path),
            (hashtype+'sum', hash),
        ])
        self.add_upload_certs(c)

        try:
            c.perform()
            status = c.getinfo(pycurl.RESPONSE_CODE)
            output = buf.getvalue().decode('utf-8').strip()
        except Exception as e:
            raise RpkgError(e)
        finally:
            c.close()

        # Render new-line after displaying the upload progress
        sys.stdout.write('\n')
        sys.stdout.flush()

        if status != 200:
            log.error(output)
            raise RpkgError('Server returned status code %d.' % status)
        else:
            log.debug(output)

    ######################## commands ########################

    def sources(self, repo_path, outdir=None, raise_on_missing_sources=True):
        if not os.path.exists(self.sources_filepath):
            if raise_on_missing_sources:
                raise RpkgError('No sources file found.')
            log.debug('No sources file found.')
            return

        if not outdir:
            outdir = self.pkg_path

        for entry in self.read_source_entries():
            outfile = os.path.join(
                outdir, entry.filename)

            self.download(outfile,
                          repo_path,
                          entry.filename,
                          entry.hashtype,
                          entry.hash)

    def upload(self, repo_path, files, offline=False):
        if not repo_path and not offline:
            raise RpkgError('repo_path cannot be empty.')

        source_entries = []
        for infile in files:
            entry = self.make_source_entry(infile)
            if not offline:
                self.upload_source(infile,
                                   repo_path,
                                   entry.filename,
                                   entry.hashtype,
                                   entry.hash)
            source_entries.append(entry)

        if not offline:
            log.info('Source upload finished.')
        else:
            log.info('Source upload skipped. Please, upload them later.')

        return source_entries
